def conversations_to_messages(sample: dict) -> list:
    """
    Simple function to convert a conversation to a list of messages.

    Example:
        {"conversation": [{"from": "human", "value": "human text"}, {"from": "bot", "value": "bot text"}]}

    Args:
        sample (dict): a conversation dictionary object

    Returns:
        list: a list of messages compatible with the OpenAI format
        [{"role": "user", "content": "human text"}, {"role": "assistant", "content": "bot text"}]
    """
    messages = []

    # Add system message if present
    if 'system' in sample:
        messages.append({
            "role": "system",
            "content": sample['system']
        })

    # Add conversation messages, fix roles
    for item in sample['conversations']:
        if item['from'] == 'user':
            role = 'user'
        elif item['from'] == 'human':
            role = 'user'
        elif item['from'] == 'gpt':
            role = 'assistant'
        elif item['from'] == 'bot':
            role = 'assistant'
        elif item['from'] == 'system':
            role = 'system'
        else:
            raise ValueError(f"Invalid 'from' value: {item['from']}")
        messages.append({
            "role": role,
            "content": item['value']
        })
    return messages



import json
from typing import List, Dict, Callable, Optional
from tqdm import tqdm

import torch
from torch.utils.data import Dataset
from transformers import AutoTokenizer, logging

from impruver.data import apply_chat_template

_log = logging.get_logger()
logging.set_verbosity_info()


class ChatDataset(Dataset):
    def __init__(
        self,
        original_records: List[Dict],
        tokenizer,
        max_tokens_count: int,
        add_global_bos: bool = True,
        add_global_eos: bool = True,
        labels_pad_token_id: int = -100,
        converter: Optional[Callable[[Dict], List[Dict]]] = None,
        strategy_function: Optional[Callable[[List[Dict]], List[Dict]]] = None,
        chat_template: Optional[str] = None,
        only_target_loss: bool = False,
        mapping: Optional[Dict[str, str]] = None,
    ):
        self.original_records = original_records
        self.tokenizer = tokenizer
        self.max_tokens_count = max_tokens_count
        self.labels_pad_token_id = labels_pad_token_id
        self.add_global_bos = add_global_bos
        self.add_global_eos = add_global_eos
        self.converter = converter
        self.strategy_function = strategy_function
        self.chat_template = chat_template
        self.only_target_loss = only_target_loss
        self.mapping = mapping
        self.is_printed = False

        self.records = []
        for record in tqdm(original_records):
            tensors = self.convert_record(record)
            if tensors is None:
                continue
            self.records.append(tensors)

    def __len__(self):
        return len(self.records)

    def __getitem__(self, index):
        return self.records[index]

    def get_tokens(self, messages) -> List[int]:
        if (
            hasattr(self.tokenizer, 'apply_chat_template')
            and hasattr(self.tokenizer, 'chat_template')
            and self.tokenizer.chat_template is not None
        ):
            tokens = self.tokenizer.apply_chat_template(
                messages,
                chat_template=self.chat_template,
                add_special_tokens=False,
                tokenize=True,
                add_generation_prompt=False,
            )
        else:
            # Assume apply_chat_template is a function you have defined
            tokens = apply_chat_template(
                messages,
                chat_template=self.chat_template,
                add_special_tokens=False,
                tokenize=True,
                add_generation_prompt=False,
                tokenizer=self.tokenizer,
            )

        if not isinstance(tokens, list):
            return tokens.tolist()[0]
        return tokens

    def convert_record(self, record):
        if self.converter:
            if self.mapping is not None:
                messages = self.converter(record, mapping=self.mapping)
            else:
                messages = self.converter(record)
        else:
            messages = record["messages"]
            if isinstance(record["messages"], str):
                messages = json.loads(record["messages"])

        # Ig sample is empty, then we don't need it
        if messages is None:
            return None

        if self.strategy_function:
            messages = self.strategy_function(
                messages=messages,
                tokenizer=self.tokenizer,
                max_tokens_count=self.max_tokens_count,
                chat_template=self.chat_template,
            )

        # Tokenize the entire conversation
        input_ids = self.get_tokens(messages)
        if len(input_ids) > self.max_tokens_count - 2:
            # If length of conversation is larger than allowed then skip this sample
            _log.info(f'Input is "{len(input_ids)}" tokens, max allowed is "{self.max_tokens_count}" tokens, skip...')
            return None

        # If only_target_loss=False, then labels is a copy of input_ids
        if self.only_target_loss:
            labels = []
            for message in messages:
                message_tokens = self.get_tokens([message])
                if message["role"] in ("assistant", "function_call"):
                    labels += message_tokens
                else:
                    labels += [self.labels_pad_token_id] * len(message_tokens)
        else:
            labels = input_ids

        # Add global BOS and EOS tokens if specified
        if self.add_global_bos and input_ids[0] != self.tokenizer.bos_token_id:
            input_ids = torch.cat(([self.tokenizer.bos_token_id], input_ids), dim=-1)
            labels = [self.labels_pad_token_id] + labels

        if self.add_global_eos and input_ids[-1] != self.tokenizer.eos_token_id:
            input_ids = torch.cat((input_ids, [self.tokenizer.eos_token_id]), dim=-1)
            labels += [self.tokenizer.eos_token_id]

        # Convert to tensors
        input_ids = torch.LongTensor(input_ids)
        labels = torch.LongTensor(labels)
        attention_mask = torch.ones_like(input_ids)

        # Ensure lengths are consistent
        assert (
            input_ids.size(0)
            == labels.size(0)
            == attention_mask.size(0)
            <= self.max_tokens_count
        )

        # Print sample once for verification
        if not self.is_printed:
            print("Input IDs:", input_ids)
            print("Labels:", labels)
            print("Full prompt:", self.tokenizer.decode(input_ids.tolist(), skip_special_tokens=False))
            self.is_printed = True

        return {
            "input_ids": input_ids.tolist(),
            "labels": labels.tolist(),
            "attention_mask": attention_mask.tolist(),
        }


if __name__ == '__main__':
    records = [
        {
            "messages": [
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Hello, how are you?"},
                {"role": "assistant", "content": "I'm doing great!"},
                {"role": "user", "content": "Can you help me with something?"},
                {"role": "assistant", "content": "Sure, what do you need help with?"}
            ]
        }
    ]
    tokenizer = AutoTokenizer.from_pretrained('gpt2')
    dataset = ChatDataset(
        original_records=records,
        tokenizer=tokenizer,
        max_tokens_count=1024,
        only_target_loss=True,
    )

    print(dataset[0])



import argparse
import os
import runpy
import sys
import textwrap

from pathlib import Path
from typing import Optional

import impruver

from torch.distributed.elastic.multiprocessing.errors import record
from torch.distributed.run import get_args_parser as get_torchrun_args_parser, run
from impruver._cli.subcommand import Subcommand
from impruver._recipe_registry import Config, get_all_recipes, Recipe

ROOT = Path(impruver.__file__).parent.parent


class Run(Subcommand):
    """
    Содержит в себе всю логику для запуска подкоманд через `impriver run`.
    """

    def __init__(self, subparsers):
        super().__init__()
        self._parser = subparsers.add_parser(
            "run",
            prog="impruver run",
            help="Run a recipe. For distributed recipes, this supports all torchrun arguments.",
            description="Run a recipe. For distributed recipes, this supports all torchrun arguments.",
            usage="impruver run [TORCHRUN-OPTIONS] <recipe> --config <config> [RECEIPT-OPTIONS]",
            epilog=textwrap.dedent(
                """\
                examples:

                    # Run a finetuning recipe on a single device w/ default values
                    $ impruver run \
                        finetune ruGPT-3.5/13B_lora_saiga2

                    # Run a finetuning recipe in a distributed fashion using torchrun and specify a number of GPUs for torchrun
                    $ impruver run \
                        --nproc_per_node 4 \
                        finetune ruGPT-3.5/13B_lora_saiga2

                Remember, you can use `impruver cp` to copy a default recipe/config to your local dir and modify the values.
                """
            ),
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self._add_arguments()
        self._parser.set_defaults(func=self._run_cmd)

    def _add_arguments(self) -> None:
        """
        Add arguments to the parser.

        This is a bit hacky since we need to add the torchrun arguments to our parser.
        This grabs the argparser from torchrun, iterates over it's actions, and adds them
        to our parser. We rename the training_script and training_script_args to recipe and recipe_args
        respectively. In addition, we leave out the help argument since we add it manually to ours.
        """
        torchrun_argparser = get_torchrun_args_parser()
        for action in torchrun_argparser._actions:
            if action.dest == "training_script":
                action.dest = "recipe"
                action.help = ("Name or path to recipe to be launched followed by args. "
                               "For a list of all possible recipes, run `impruver ls`.")
            elif action.dest == "training_script_args":
                action.dest = "recipe_args"
                action.help = "Args to be passed to the recipe."
            elif action.dest == "help":
                continue
            self._parser._add_action(action)

    @record
    def _run_distributed(self, args: argparse.Namespace, is_builtin: bool):
        """
        Run a recipe with torchrun.
        """
        # we don't do this since we test on CPUs for distributed. Will update once multi GPU CI is supported.
        print("Running with torchrun...")
        # Have to reset the argv so that the recipe can be run with the correct arguments
        args.training_script = args.recipe
        args.training_script_args = args.recipe_args

        # If the user does not explicitly pass a rendezvous endpoint, run in standalone mode.
        # This allows running multiple distributed training jobs simultaneously.
        if not args.rdzv_endpoint:
            args.standalone = True

        # impruver built-in recipes are specified with an absolute posix path, but
        # custom recipes are specified as a relative module dot path and need to be
        # run with python -m
        args.module = not is_builtin
        run(args)

    def _run_single_device(self, args: argparse.Namespace, is_builtin: bool):
        """Run a recipe on a single device."""
        sys.argv = [str(args.recipe)] + args.recipe_args
        if is_builtin:
            # impruver built-in recipes are specified with an absolute posix path
            runpy.run_path(str(args.recipe), run_name="__main__")
        else:
            # custom recipes are specified as a relative module dot path
            runpy.run_module(str(args.recipe), run_name="__main__")

    def _is_distributed_args(self, args: argparse.Namespace):
        """Check if the user is trying to run a distributed recipe."""
        total = len(sys.argv) - 2  # total args minus "tune run"
        script_args = len(args.recipe_args) + 1  # script args + 1 for script name
        return total > script_args

    def _get_recipe(self, recipe_str: str) -> Optional[Recipe]:
        """Get a recipe from the name or path.

        Args:
            recipe_str (str): The name or path of the recipe.

        Returns:
            The recipe if it's found in built-in recipes, otherwise None.
        """
        for recipe in get_all_recipes():
            if recipe.name == recipe_str:
                return recipe

    def _convert_to_dotpath(self, recipe_path: str) -> str:
        """Convert a custom recipe path to a dot path that can be run as a module.

        Args:
            recipe_path (str): The path of the recipe.

        Returns:
            The dot path of the recipe.
        """
        filepath, _ = os.path.splitext(recipe_path)
        return filepath.replace("/", ".")

    def _get_config(
        self, config_str: str, specific_recipe: Optional[Recipe]
    ) -> Optional[Config]:
        """Get a config from the name or path.

        Args:
            config_str (str): The name or path of the config.
            specific_recipe (Optional[Recipe]): The specific recipe to search through.

        Returns:
            The config if it's found in built-in configs, otherwise None.
        """
        # If a specific recipe is provided, search through it
        if specific_recipe is not None:
            for config in specific_recipe.configs:
                if config.name == config_str:
                    return config

        # If not, search through all recipes
        for recipe in get_all_recipes():
            for config in recipe.configs:
                if config.name == config_str:
                    return config

    def _run_cmd(self, args: argparse.Namespace):
        """Run a recipe."""
        # We have to assume that the recipe supports distributed training
        supports_distributed = True
        recipe_path, config_path = None, None

        # Try to find config string in args
        try:
            config_idx = args.recipe_args.index("--config") + 1
            config_str = args.recipe_args[config_idx]
        except ValueError:
            self._parser.error("The '--config' argument is required.")

        # Get recipe path
        recipe = self._get_recipe(args.recipe)
        if recipe is None:
            recipe_path = self._convert_to_dotpath(args.recipe)
            is_builtin = False
        else:
            recipe_path = str(ROOT / "recipes" / recipe.file_path)
            supports_distributed = recipe.supports_distributed
            is_builtin = True

        # Get config path
        config = self._get_config(config_str, recipe)
        if config is None:
            config_path = config_str
        else:
            config_path = str(ROOT / "recipes" / "configs" / config.file_path)

        # Prepare args
        args.recipe = recipe_path
        args.recipe_args[config_idx] = config_path

        # Make sure user code in current directory is importable
        sys.path.append(os.getcwd())

        # Execute recipe
        if self._is_distributed_args(args):
            if not supports_distributed:
                self._parser.error(
                    f"Recipe {recipe.name} does not support distributed training."
                    "Please run without torchrun commands."
                )
            self._run_distributed(args, is_builtin=is_builtin)
        else:
            self._run_single_device(args, is_builtin=is_builtin)



from .instruction_to_messages import instruction_to_messages
from .dialog_to_messages import dialog_to_messages
from .conversations_to_messages import conversations_to_messages
from .char_dialog_to_messages import char_dialog_to_messages



import importlib


def dynamic_import(module_class_string: str, class_name: str = None, default_class: str = None):
    try:
        module_name, extracted_class_name = module_class_string.rsplit(".", 1)
        class_name = class_name or extracted_class_name
        module = importlib.import_module(module_name)

        if hasattr(module, class_name):
            cls = getattr(module, class_name)
        elif default_class and hasattr(module, default_class):
            cls = getattr(module, default_class)
        else:
            raise ImportError(f"Class {class_name} or default class {default_class} not found in module {module_name}")

        return cls
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Could not import {module_class_string}: {e}")



import logging
from typing import List, Optional
from impruver.data.message import Message
from impruver.data.tokenizer import Tokenizer
from impruver.data.apply_chat_template import apply_chat_template
from impruver.utils.get_logger import get_logger

_log: logging.Logger = get_logger()


def last_message_by_assistant(
        tokenizer: Tokenizer,
        messages: List[Message],
        max_tokens_count: int,
        chat_template: Optional[str] = None,
) -> List[Message]:
    """
    Return the last message by assistant.

    If there is no message by assistant, return the first message.
    If there is more than one message by assistant, return the last one.

    Args:
        tokenizer (Tokenizer): Tokenizer object
        messages (List[Message]): List of Message objects
        max_tokens_count (int): Max tokens count
        chat_template (Optional[str]): The Jinja2 template to apply to the conversation. Defaults to None.

    Returns:
        List[Message]: List of Message objects
    """

    _log.debug(f"Max tokens count: {max_tokens_count}")
    tmp_messages = messages.copy()
    while tmp_messages:

        # Apply chat format from template
        if hasattr(tokenizer, 'apply_chat_template'):
            # On modern tokenizers we may use chat_template
            formated_messages = tokenizer.apply_chat_template(
                tmp_messages,
                chat_template=chat_template,  # Use default_chat_template if None
                tokenize=False
            )
        else:
            # On old tokenizers we will use a custom apply_chat_template
            formated_messages = apply_chat_template(
                tmp_messages,
                chat_template=chat_template,  # Use DEFAULT_CHAT_TEMPLATE if None
                tokenize=False,
                tokenizer=tokenizer
            )

        # Tokenize all messages
        tokenized_messages = tokenizer.encode(formated_messages, return_tensors="pt")

        # Calculate sum of total tokens of all formated messages
        total_tokens = sum(len(tokens) for tokens in tokenized_messages)
        _log.debug(f"Total tokens: {total_tokens}")

        # If total tokens pass the limit
        if total_tokens <= max_tokens_count:
            # If last message is from the assistant
            if tmp_messages[-1].role == 'assistant':
                # Return the messages, formated and tokenized they versions
                return tmp_messages
            else:
                # Remove the last message, then try again
                tmp_messages.pop()
        else:
            # Remove the last message if total tokens exceed the limit
            tmp_messages.pop()

    _log.error(
        "Unable to fit messages within the maximum token length "
        "with the last message being from the assistant."
    )

    return messages



import logging
import os
import random
from typing import Optional, Union
import numpy as np
import torch

from .get_logger import get_logger

_log: logging.Logger = get_logger()


def set_seed(
        seed: Optional[int] = None,
        debug_mode: Optional[Union[str, int]] = None
) -> int:
    if seed is None:
        seed = random.randint(1, 2 ** 32 - 1)

    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

    if debug_mode is not None:
        _log.debug(f"Setting deterministic debug mode to {debug_mode}")
        torch.set_deterministic_debug_mode(debug_mode)
        deterministic_debug_mode = torch.get_deterministic_debug_mode()
        if deterministic_debug_mode == 0:
            _log.debug("Disabling cuDNN deterministic mode")
            torch.backends.cudnn.deterministic = False
            torch.backends.cudnn.benchmark = True
        else:
            _log.debug("Enabling cuDNN deterministic mode")
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
            os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

    return seed



from typing import List, Protocol, Set, Dict


class Tokenizer(Protocol):
    """Abstract tokenizer"""

    bos_token_id: int
    eos_token_id: int
    pad_token_id: int

    bos_token: str
    eos_token: str
    pad_token: str

    # Tokens indicating that generation should stop
    stop_tokens: Set[int]

    # Max length of model' context
    model_max_length: int

    def encode(self, text: str | List[str], **kwargs) -> List[int] | List[List[int]]:
        """
        Given a string, return a list of token ids
        """

    def decode(self, token_ids: List[int], add_bos: bool, add_eos: bool, **kwargs) -> str:
        """
        Given a list of token ids, return the decoded text
        """

    def apply_chat_template(self, token_ids: List[Dict], **kwargs):
        """
        Given a list of messages, return a list of tokens for the concatenated
        and formatted message
        """
        pass



import random


def build_char_system_messages(char):
    name = char["name"]
    greeting = char["greeting"]
    example_dialogue = char["example_dialogue"]
    context = ""
    if random.random() < 0.5:
        context += f"Ты {name}. "
    context += f"{char['context']}"
    chat = []
    if random.random() < 0.2:
        context += f"\nПриветствие: {greeting}"
        chat.append({"role": "char", "content": greeting})
    if random.random() < 0.2:
        mapping = {"user": "Пользователь", "char": "Персонаж"}
        example_messages = [f'{mapping[m["role"]]}: {m["content"]}' for m in example_dialogue]
        context += "\nПример диалога:\n" + "\n".join(example_messages)
    chat.insert(0, {"role": "system", "content": context})
    return chat


def char_dialog_to_messages(sample: dict) -> list:
    messages_list = []
    for dialogue in sample["dialogues"]:
        if dialogue["model_name"] != "gpt-4":
            continue
        chat = dialogue["chat"]
        for message in chat:
            if message["role"] == "char":
                message["role"] = "assistant"
            if message["role"] == "operator" or message["role"] == "user":
                message["role"] = "user"
        system_messages = build_char_system_messages(sample)
        messages = system_messages + chat
        messages_list.append(messages)
    return messages_list



from typing import List, Dict

from impruver.data.tokenizer import Tokenizer


def get_prompt(tokenizer: Tokenizer, messages: List[Dict], add_generation_prompt: bool = False):
    return tokenizer.apply_chat_template(
        messages,
        add_special_tokens=False,
        tokenize=False,
        add_generation_prompt=add_generation_prompt,
    )



import os
from dataclasses import dataclass
from typing import List
from pathlib import Path

import impruver

ROOT = Path(impruver.__file__).parent.parent


@dataclass
class Config:
    """
    Описание одного конкретного набора настроек (yaml-файл).
    """
    name: str
    file_path: str


@dataclass
class Recipe:
    """
    Объект описывающий рецепт запуска

    :param name: Название рецепта (например, "full_finetune")
    :param file_path: Путь до исполняемого скрипта (здесь "finetune_transformers.py")
    :param configs: Список объектов Config, содержащих конфигурации
    :param supports_distributed: Флаг, указывающий, поддерживает ли рецепт распределённый запуск
    """
    name: str
    file_path: str
    configs: List[Config]
    supports_distributed: bool


def get_all_recipes(config_dir: str = ROOT / "recipes") -> List[Recipe]:
    all_configs: List[Config] = []
    for root, test, filenames in os.walk(config_dir):
        sorted_filenames = sorted(filenames)
        for filename in sorted_filenames:
            if not filename.endswith(".yaml"):
                continue
            file_path = os.path.join(root, filename)
            folder_name = os.path.basename(root)
            file_base = os.path.splitext(filename)[0]
            config_obj = Config(name=folder_name + '/' + file_base, file_path=file_path)
            if config_obj is None:
                continue
            all_configs.append(config_obj)
    # Сортируем все конфигурации в алфавитном порядке по атрибуту file_path
    all_configs.sort(key=lambda x: x.file_path.lower())

    recipe_compose_dataset = Recipe(
        name="compose_dataset",
        file_path="compose_dataset.py",
        configs=[],
        supports_distributed=False,
    )
    recipe_chat = Recipe(
        name="chat",
        file_path="chat_transformers.py",
        configs=[],
        supports_distributed=False,
    )
    recipe_finetune_transformers = Recipe(
        name="finetune",
        file_path="finetune_transformers.py",
        configs=all_configs,
        supports_distributed=True,
    )
    recipe_convert_gguf = Recipe(
        name="convert_gguf",
        file_path="convert_gguf.py",
        configs=[],
        supports_distributed=False,
    )
    recipe_finetune_unsloth = Recipe(
        name="unsloth",
        file_path="finetune_unsloth.py",
        configs=[],
        supports_distributed=True,
    )

    return [
        recipe_finetune_transformers,
        recipe_finetune_unsloth,
        recipe_compose_dataset,
        recipe_chat,
        recipe_convert_gguf
    ]


if __name__ == "__main__":
    import recipes

    recipes_files = get_all_recipes(os.path.join(recipes.__path__[0], "/configs"))
    for recipe in recipes_files:
        print(f"\nRecipe: {recipe.name}")
        print(f"  Script: {recipe.file_path}")
        print(f"  Distributed: {recipe.supports_distributed}")
        print("  Configs:")
        for conf in recipe.configs:
            print(f"    - {conf.name}  ->  {conf.file_path}")
    print()



from typing import List

from .message import Message


def validate_messages(messages: List[Message]) -> None:
    """
    Given a list of messages, ensure that messages form a valid
    back-and-forth conversation. An error will be raised if:

    - There is a system message that's not the first message
    - There are two consecutive user messages
    - An assistant message comes before the first user message
    - The message is empty
    - Messages are shorter than length of 2 (min. one user-assistant turn)


    Args:
        messages (List[Message]): the messages to validate.

    Raises:
        ValueError: If the messages are invalid.
    """
    if len(messages) < 2:
        raise ValueError(
            f"Messages must be at least length 2, but got {len(messages)} messages"
        )

    last_turn = "assistant"
    for i, message in enumerate(messages):
        if message.role == "assistant" and last_turn != "user" and last_turn != "system":
            raise ValueError(
                f"Assistant message before expected user message at index {i} in messages"
            )
        if message.role == "user" and last_turn == "user":
            raise ValueError(
                f"Two consecutive user messages at index {i} and {i - 1} in messages"
            )
        if message.role == "system" and i > 0:
            raise ValueError(
                f"System message at index {i} in messages, but system messages must come first"
            )
        last_turn = message.role



import json


def read_jsonl(file_name):
    with open(file_name, encoding="utf-8") as r:
        return [json.loads(line) for line in r]


def write_jsonl(records, path):
    with open(path, "w", encoding="utf-8") as w:
        for r in records:
            w.write(json.dumps(r, ensure_ascii=False) + "\n")



import os
import subprocess
from typing import Dict, List

import fire
import yaml
from peft import AutoPeftModelForCausalLM

from impruver.utils import dynamic_import


def convert_gguf(
    config: str,
    llama_cpp_dir: str = "../llama.cpp",
    llama_cpp_quantize_bin: str = "../llama.cpp/llama-quantize",
    quantizations: List[str] = ["q8_0", "q5_0", "q4_0", "q2_k"],
):
    """
    Convert a PyTorch model to a GGUF format, if provider mode is LoRA adapter,
    then merge if first, then convert. After that, quantize the model to list of required levels
    and save nearby a Modelfiles with configuration for importing models to Ollama.

    Args:
        config (str): Path to the configuration file
        llama_cpp_dir (str): Path to the llama.cpp directory
        llama_cpp_quantize_bin (str): Path to the llama-quantize binary
        quantizations (List[str]): List of quantizations to use
    """

    #
    # Load configuration
    #

    if os.path.exists(config):
        config_path = config
    else:
        import recipes
        recipes_path = os.path.join(recipes.__path__[0])
        config_path = recipes_path + '/configs/' + config + '.yaml'

    # Read config
    with open(config_path, "r") as r:
        config = yaml.safe_load(r)

    # Output dir for merged and converted models
    output_dir = config["output_dir"]
    processing_dir = output_dir + "/processing"
    os.makedirs(processing_dir, exist_ok=True)
    gguf_dir = output_dir + "/gguf"
    os.makedirs(gguf_dir, exist_ok=True)

    # Class to work with Tokenizer
    tokenizer_class = "transformers.AutoTokenizer"
    if "class" in config["tokenizer"]:
        tokenizer_class = config["tokenizer"]["class"]

    #
    # Model preparation
    #

    # If LoRA is enabled, we assume we have a LoRA adapter in output_dir
    if config.get("lora", None):
        # This mean we need to merge an adapter into a model then save result to disk
        peft_model = AutoPeftModelForCausalLM.from_pretrained(output_dir, device_map={"": "cpu"}).to("cpu")
        model_processed = peft_model.merge_and_unload()
        model_processed.save_pretrained(processing_dir)

        # Next need to save tokenizer into the same folder
        tokenizer_obj = dynamic_import(tokenizer_class)
        tokenizer = tokenizer_obj.from_pretrained(output_dir)
        tokenizer.save_pretrained(processing_dir)
    else:
        processing_dir = output_dir

    #
    # Convert to GGUF format
    #

    gguf_model_path = os.path.join(gguf_dir, "model-fp16.gguf")
    convert_script = os.path.join(llama_cpp_dir, "convert_hf_to_gguf.py")

    print("Converting model to GGUF (FP16)...")
    convert_cmd = ["python", convert_script, "--outtype", "f16", "--outfile", gguf_model_path, processing_dir]
    subprocess.run(convert_cmd, check=True)

    # Create the Modelfile.f16
    q_level = "f16"
    header_file_name = f"Modelfile.{q_level}"
    header_file_path = os.path.join(gguf_dir, header_file_name)
    with open(header_file_path, "w", encoding="utf-8") as hf:
        hf.write(f"FROM model-{q_level}.gguf\n")
        hf.write("PARAMETER temperature 1\n")
        hf.write("# PARAMETER num_ctx 4096\n")
        hf.write("# SYSTEM You are Super King, acting as a king.\n")

    #
    # Quantization loop
    #

    quant_list = [q for q in quantizations]
    for q_level in quant_list:
        print(f"Quantizing to {q_level}...")
        quant_out_path = os.path.join(gguf_dir, f"model-{q_level}.gguf")
        quant_cmd = [llama_cpp_quantize_bin, gguf_model_path, quant_out_path, q_level]
        subprocess.run(quant_cmd, check=True)

        # Create the Modelfile.qX_Y
        header_file_name = f"Modelfile.{q_level}"
        header_file_path = os.path.join(gguf_dir, header_file_name)
        with open(header_file_path, "w", encoding="utf-8") as hf:
            hf.write(f"FROM model-{q_level}.gguf\n")
            hf.write("PARAMETER temperature 1\n")
            hf.write("# PARAMETER num_ctx 4096\n")
            hf.write("# SYSTEM You are Super King, acting as a king.\n")


if __name__ == "__main__":
    fire.Fire(convert_gguf)



import yaml
import fire

import torch
from transformers import logging, GenerationConfig, BitsAndBytesConfig

from impruver.utils import set_seed, get_dtype, dynamic_import
from impruver.conversation import get_prompt, generate, ChatHistory


def chat(
    config: str,
    output_dir: str | None = None,
    history_limit: int = 10,
    system_prompt: str | None = None,
    seed: int = 42,
    max_new_tokens: int = 200,
    repetition_penalty: float = 1.2,
    do_sample: bool = True,
    temperature: float = 0.5,
    top_p: float = 0.6,
    top_k: int = 40,
):
    """
    Chat with a trained model.

    Args:
        config: Path to config file
        output_dir: Path to output directory
        history_limit: Number of previous messages to keep in memory
        system_prompt: Prompt to use for the system
        seed: Seed for random number generator
        max_new_tokens: Maximum number of new tokens to generate
        repetition_penalty: Penalty for repeated tokens
        do_sample: Whether to use sampling or greedy decoding
        temperature: Temperature for sampling
        top_p: Top-p for sampling
        top_k: Top-k for sampling
    """

    set_seed(seed)
    logging.set_verbosity_info()

    #
    # Load configuration
    #

    # Read config from disk
    with open(config, "r") as r:
        config = yaml.safe_load(r)

    # Path where model will be saved
    output_dir = output_dir
    if output_dir is None:
        output_dir = config['output_dir']

    # Get settings of Peft/LoRA adapter
    lora_config = config.get("lora")

    # Class to work with Tokenizer
    model_class = "transformers.AutoModelForCausalLM"
    if "class" in config["model"]:
        model_class = config["model"]["class"]

    # Class to work with Tokenizer
    tokenizer_class = "transformers.AutoTokenizer"
    if "class" in config["tokenizer"]:
        tokenizer_class = config["tokenizer"]["class"]

    # Settings related to bitsandbytes and useful only with LoRA adapter training
    load_in_4bit = False
    if "load_in_4bit" in config["model"]:
        load_in_4bit = bool(config["model"]["load_in_4bit"])
    load_in_8bit = False
    if "load_in_8bit" in config["model"]:
        load_in_8bit = bool(config["model"]["load_in_8bit"])

    # Get ddp settings from config if available
    ddp_config = config.get("ddp", {})

    #
    # Tokenizer preparation
    #

    # Init tokenizer object
    tokenizer_obj = dynamic_import(tokenizer_class)
    tokenizer = tokenizer_obj.from_pretrained(output_dir)

    #
    # Model preparation
    #

    # Data Type is bfloat16 by default
    dtype = torch.bfloat16
    if "dtype" in config["model"]:
        dtype = get_dtype(config["model"]["dtype"])

    # Quantization related settings, can be used only in combination with LoRA adapter
    quantization_config = None
    if load_in_4bit:
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_compute_dtype=dtype,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=True
        )
    if load_in_8bit:
        quantization_config = BitsAndBytesConfig(
            load_in_8bit=True,
            llm_int8_threshold=6.0,
        )

    # Generator config
    generation_config = GenerationConfig.from_pretrained(output_dir)
    generation_config.max_new_tokens = max_new_tokens
    generation_config.repetition_penalty = repetition_penalty
    generation_config.do_sample = do_sample
    generation_config.temperature = temperature
    generation_config.top_p = top_p
    generation_config.top_k = top_k

    # Attention implementation
    attn_implementation = None
    if "attn_implementation" in config["model"]:
        attn_implementation = config["model"]["attn_implementation"]

    # Init model object
    model_obj = dynamic_import(model_class)

    # Read model from folder with trained checkpoints
    model = model_obj.from_pretrained(
        output_dir,
        generation_config=generation_config,
        quantization_config=quantization_config,
        device_map=None if ddp_config else "auto",  # need to be disabled for DDP
        torch_dtype=dtype,
        attn_implementation=attn_implementation
    )

    # If we've trained a LoRA adapter
    if lora_config:
        from peft import PeftModel
        model = PeftModel.from_pretrained(
            model=model,
            model_id=output_dir,
            torch_dtype=dtype,
        )

    #
    # Chat loop
    #

    # Start chat loop
    chat_history = ChatHistory(history_limit, system_prompt)
    while True:
        user_message = input("User: ")

        # Reset chat command
        if user_message.strip() == "/reset":
            chat_history = ChatHistory(history_limit, system_prompt)
            print("History reset completed!")
            continue

        # Skip empty messages from user
        if user_message.strip() == "":
            continue

        # Add user message to chat history
        chat_history.add_user_message(user_message)

        # Get list of messages
        prompt = get_prompt(tokenizer, chat_history.get_messages(), True)

        # Generate response
        output = generate(model, tokenizer, prompt, generation_config)

        # Save response to chat history as assistant's message
        chat_history.add_assistant_message(output)
        print("Assistant:", output)


if __name__ == "__main__":
    fire.Fire(chat)



import yaml
import json
import random
import fire

from datasets import load_dataset
from transformers import AutoTokenizer
from datasketch import MinHash, MinHashLSH, LeanMinHash
from tqdm import tqdm

from impruver.dataset import ChatDataset
from impruver.utils import dynamic_import


def calc_fingerprint(tokens, num_perm=128):
    """
    Calculate the MinHash fingerprint of a given list of token IDs
    """
    m = MinHash(num_perm=num_perm)
    for token_id in tokens:
        # Convert the token_id to bytes and update the MinHash
        m.update(token_id.to_bytes(8, byteorder='little', signed=False))
    return LeanMinHash(m)


def load_datasets(config, tokenizer, max_tokens_count):
    """
    Load datasets specified in the config and return a list of records.
    """
    all_records = []
    for dataset in config['datasets']:
        print(dataset['name'])

        # If split was specified in the config, then use it
        split = dataset.get('split', 'train')

        # If subset (called "name" in load_dataset) was specified in the config, then use it, otherwise use None
        subset = dataset.get('subset', None)

        # Some datasets have a converter function
        converter = None
        if 'converter' in dataset:
            converter = dynamic_import(dataset['converter'])

        # Load the actual dataset from HuggingFace's datasets library.
        hf_dataset = load_dataset(dataset['name'], name=subset, split=split, trust_remote_code=True)

        # Create an instance of ChatDataset
        chat_dataset = ChatDataset(
            original_records=list(hf_dataset),
            tokenizer=tokenizer,
            converter=converter,
            mapping=dataset.get("mapping", None),
            add_global_bos=dataset.get("add_global_bos", True),
            add_global_eos=dataset.get("add_global_eos", True),
            only_target_loss=dataset.get("only_target_loss", True),
            max_tokens_count=dataset.get("max_tokens_count", max_tokens_count),
        )

        # Add records to the list
        all_records.extend(chat_dataset)

        # Optionally clear the chat_dataset to free memory
        del chat_dataset

    return all_records


def deduplicate_records(records, num_perm: int = 128, threshold: float = 0.8):
    idx_to_minhash = {}

    # Step 1: Calculate MinHash for each record
    print("Fingerprinting...")
    for idx, record in tqdm(enumerate(records), total=len(records)):
        tokens = record['input_ids']

        # Calculate MinHash fingerprint directly from tokens
        minhash = calc_fingerprint(tokens, num_perm=num_perm)
        idx_to_minhash[idx] = minhash

    # Step 2: Deduplicate records using MinHashLSH
    lsh = MinHashLSH(threshold=threshold, num_perm=num_perm)
    filtered_records = []

    print("Deduplicating...")
    for idx, minhash in tqdm(idx_to_minhash.items()):
        is_dup = False
        for other_idx in lsh.query(minhash):
            other_minhash = idx_to_minhash[other_idx]
            if minhash.jaccard(other_minhash) > threshold:
                is_dup = True
                break  # Stop checking if a duplicate is found
        if not is_dup:
            lsh.insert(idx, minhash)
            filtered_records.append(records[idx])

    return filtered_records


def split_and_save_records(records: list, train_path: str, val_path: str):
    """
    Split records into train and validation sets and save them to files.
    """
    random.shuffle(records)
    border = int(0.95 * len(records))
    train_records = records[:border]
    val_records = records[border:]

    # Write the records to files
    with open(train_path, "w", encoding='utf-8') as w:
        for record in train_records:
            w.write(json.dumps(record, ensure_ascii=False).strip() + "\n")

    with open(val_path, "w", encoding='utf-8') as w:
        for record in val_records:
            w.write(json.dumps(record, ensure_ascii=False).strip() + "\n")


def compose_dataset(
    config: str,
    train_path: str = None,
    val_path: str = None
):
    """
    Compose a dataset from multiple datasets specified in the config file.

    Args:
        config (str): Path to the config file
        train_path (str): Path to the train set
        val_path (str): Path to the validation set
    """

    # Load the config file
    with open(config, "r") as r:
        config = yaml.safe_load(r)

    # Get paths to train and validation sets
    if train_path is None:
        train_path = config['train_path']
    if val_path is None:
        val_path = config['val_path']

    # Get tokenizer
    tokenizer = AutoTokenizer.from_pretrained(config['tokenizer']['name'])

    # Use provided chat_template if set in config
    if 'chat_template' in config['tokenizer']:
        tokenizer.chat_template = config['tokenizer']['chat_template']

    # Default max_tokens_count from tokenizer
    max_tokens_count = tokenizer.model_max_length

    # If max_tokens_count was replaced in config, then overwrite it
    if 'max_tokens_count' in config['tokenizer']:
        max_tokens_count = config['tokenizer']['max_tokens_count']

    # Prevent automatic truncation by setting model_max_length to a large value
    tokenizer.model_max_length = int(1e30)  # Disable automatic truncation

    # Step 1: Load datasets
    records = load_datasets(config, tokenizer, max_tokens_count)

    # Step 2: Deduplicate records
    deduplicated_records = deduplicate_records(records)

    # Step 3: Split and save records
    split_and_save_records(deduplicated_records, train_path, val_path)


if __name__ == "__main__":
    fire.Fire(compose_dataset)



from .get_logger import get_logger
from .get_device import get_device
from .get_dtype import get_dtype
from .set_seed import set_seed
from .dynamic_import import dynamic_import
from .io import read_jsonl, write_jsonl



class Subcommand:
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def _add_arguments(self):
        pass



import argparse
import json
import os
import textwrap
import traceback
from pathlib import Path

from huggingface_hub import snapshot_download
from huggingface_hub.utils import GatedRepoError, RepositoryNotFoundError

from impruver._cli.subcommand import Subcommand


class Download(Subcommand):
    def __init__(self, subparsers: argparse._SubParsersAction):
        super().__init__()
        self._parser = subparsers.add_parser(
            "download",
            prog="impruver download",
            usage="impruver download <repo-id> [OPTIONS]",
            help="Download a model or a dataset from the Hugging Face Hub.",
            description="Download a model or a dataset from the Hugging Face Hub.",
            epilog=textwrap.dedent(
                """
                examples:
                    # Download a model from the HuggingFace Hub with a HuggingFace API token
                    $ tune download ai-forever/ruGPT-3.5-13B --hf-token <HF_TOKEN>

                    # Download a dataset from the HuggingFace Hub and save it to ./my_datasets/my_dataset
                    $ tune download IlyaGusev/ru_turbo_alpaca --output-dir ./my_datasets/my_dataset --repo-type=dataset
                """
            ),
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self._add_arguments()
        self._parser.set_defaults(func=self._download_cmd)

    def _add_arguments(self) -> None:
        self._parser.add_argument(
            "repo_id", type=str,
            help="Name of the repository on Hugging Face Hub (ID of model or dataset).",
        )
        self._parser.add_argument(
            "--repo-type", type=str, required=False, default="model", choices=["dataset", "model"],
            help="Repository type, can be 'dataset' or 'model'. Defaults to 'model'.",
        )
        self._parser.add_argument(
            "--output-dir", type=Path, required=False, default=None,
            help="Directory in which to save the model. Defaults to `./<repo_id>`.",
        )
        self._parser.add_argument(
            "--hf-token", type=str, required=False, default=os.getenv("HF_TOKEN", None),
            help="Hugging Face API token. Required for private repositories.",
        )
        self._parser.add_argument(
            "--ignore-patterns", type=str, required=False,
            help="If provided, files matching any of the patterns are not downloaded. Example: '*.safetensors'.",
        )

    def _download_cmd(self, args: argparse.Namespace) -> None:

        # Default output_dir is `./<model_name>`
        output_dir = args.output_dir
        if output_dir is None:
            repo_name = args.repo_id.split("/")[-1]
            output_dir = Path(".") / repo_name

        try:
            true_output_dir = snapshot_download(
                repo_id=args.repo_id,
                repo_type=args.repo_type,
                local_dir=output_dir,
                ignore_patterns=args.ignore_patterns,
                token=args.hf_token,
            )
        except GatedRepoError:
            if args.hf_token:
                self._parser.error(
                    "It looks like you are trying to access a gated repository. Please ensure you "
                    "have access to the repository."
                )
            else:
                self._parser.error(
                    "It looks like you are trying to access a gated repository. Please ensure you "
                    "have access to the repository and have provided the proper Hugging Face API token "
                    "using the option `--hf-token` or by running `huggingface-cli login`."
                    "You can find your token by visiting https://huggingface.co/settings/tokens"
                )
        except RepositoryNotFoundError:
            self._parser.error(
                f"Repository '{args.repo_id}' not found on the HuggingFace Hub."
            )
        except Exception as e:
            tb = traceback.format_exc()
            self._parser.error(
                f"Failed to download {args.repo_id} with error: '{e}' and traceback: {tb}"
            )

        # Save repo metadata
        metadata_path = output_dir / "repo_metadata.json"
        with open(metadata_path, "w") as metadata_file:
            json.dump({"repo_id": args.repo_id}, metadata_file, indent=4)

        print(
            "Successfully downloaded model repo and wrote to the following locations:",
            *list(Path(true_output_dir).iterdir()),
            sep="\n",
        )



import torch


def get_device(device: str | int) -> torch.device:
    if device == "cpu":
        return torch.device("cpu")
    elif device == "cuda":
        return torch.device("cuda")
    else:
        return torch.device(f"cuda:{device}")



from typing import Optional, List, Union, Dict
from jinja2 import Template

from .tokenizer import Tokenizer

DEFAULT_RAW_TEMPLATE = (
    "{% set loop_messages = messages %}"
    "{% for message in loop_messages %}"
        "{% set content = '' + message['content'] | trim + '' %}"
        '{{ bos_token + content + eos_token }}'
    "{% endfor %}"
)

DEFAULT_CHAT_TEMPLATE = (
    "{% set loop_messages = messages %}"
    "{% for message in loop_messages %}"
        "{% set content = '' + message['role'] + '\n' + message['content'] | trim + '' %}"
        "{{ bos_token + content + eos_token}}"
    "{% endfor %}"
    "{% if add_generation_prompt %}"
        "{% set content = 'assistant' + '\n' %}"
        "{{ bos_token + content }}"
    "{% endif %}"
)


def apply_chat_template(
        conversation: List[Dict],
        chat_template: Optional[str] = None,
        add_special_tokens: bool = False,
        add_generation_prompt: Optional[bool] = False,
        tokenize: bool = True,
        tokenizer: Tokenizer = None
) -> Union[List[int], Dict, str]:
    """
    Apply a chat template to a conversation in GPT2 sample format.

    Args:
        conversation (List[dict]): The conversation to apply the template to.
        chat_template (Optional[str]): The Jinja2 template to apply to the conversation. Defaults to None.
        add_special_tokens (bool): Whether to add begin and end of text tokens. Defaults to False.
        add_generation_prompt (Optional[bool]): Whether to add generation prompt. Defaults to False.
        tokenize (bool): Whether to tokenize the rendered text. Defaults to True.
        tokenizer (Optional[Tokenizer]): The tokenizer to use. Defaults to None.

    Returns:
        Union[List[int], Dict, str]: The rendered text.
    """

    # Use provided chat template or default template
    template_str = chat_template if chat_template else DEFAULT_CHAT_TEMPLATE
    template = Template(template_str)

    # Prepare the context for the template
    context = {
        'messages': [{'role': msg['role'], 'content': msg['content']} for msg in conversation],
        'add_generation_prompt': add_generation_prompt,
        'bos_token': tokenizer.bos_token if tokenizer else '',
        'eos_token': tokenizer.eos_token if tokenizer else ''
    }

    # Render the template
    rendered = template.render(context).strip()

    # Add begin and end of text tokens
    if add_special_tokens:
        rendered = "<|startoftext|>" + rendered + "<|endoftext|>"

    if tokenize and tokenizer:
        # Tokenize the rendered text
        return tokenizer.encode(rendered, return_tensors="pt")
    else:
        # Return the rendered text as is
        return rendered



from typing import List, Dict


class ChatHistory:
    def __init__(self, history_limit: int = None, system_prompt: str = None):
        self.history_limit: int | None = history_limit
        self.system_prompt: str | None = system_prompt
        self.messages: List[Dict] = []
        if self.system_prompt is not None:
            self.messages.append({"role": "system", "content": self.system_prompt})

    def add_message(self, role: str, message: str):
        self.messages.append({"role": role, "content": message})
        self.trim_history()

    def add_user_message(self, message: str):
        self.add_message("user", message)

    def add_assistant_message(self, message: str):
        self.add_message("assistant", message)

    def add_function_call(self, message: str):
        self.add_message("function_call", message)

    def add_function_response(self, message: str):
        self.add_message("function_response", message)

    def trim_history(self):
        appendix = 0
        if self.system_prompt is not None:
            appendix = 1
        if self.history_limit is not None and len(self.messages) > self.history_limit + appendix:
            overflow = len(self.messages) - (self.history_limit + appendix)
            self.messages = [self.messages[0]] + self.messages[overflow + appendix:]

    def get_messages(self) -> list:
        return self.messages



import os
import random
import fire
import yaml
import json

import wandb
import torch
from transformers import Trainer, logging, BitsAndBytesConfig, DataCollatorForTokenClassification
from peft import get_peft_model, LoraConfig
from impruver.utils import set_seed, read_jsonl, get_dtype, dynamic_import
from impruver.data import DEFAULT_CHAT_TEMPLATE

try:
    from unsloth import UnslothTrainingArguments
    from unsloth.trainer import _create_unsloth_optimizer
except ImportError as e:
    print(
        "`unsloth` package is not installed. Please run the following command: "
        "`wget -qO- https://raw.githubusercontent.com/unslothai/unsloth/main/unsloth/_auto_install.py | python -` "
        "to obtain command which would help you to install a correct version."
    )
    exit()


# os.environ["TOKENIZERS_PARALLELISM"] = "false"
# torch._dynamo.config.cache_size_limit = 128


class CustomTrainer(Trainer):
    def create_optimizer(self):
        embedding_learning_rate = getattr(self.args, "embedding_learning_rate", None)
        if embedding_learning_rate is None:
            return super().create_optimizer()
        if self.optimizer is None:
            optimizer_cls, optimizer_kwargs = Trainer.get_optimizer_cls_and_kwargs(self.args)
            self.optimizer = _create_unsloth_optimizer(
                self.model,
                optimizer_cls,
                optimizer_kwargs,
                embedding_learning_rate,
            )
        return self.optimizer


def finetune(
    config: str,
    train_path: str = None,
    val_path: str = None,
    output_dir: str = None,
    report_to: str = "none",
    seed: int = 42,
):
    """
    Finetune a model using classes from `unsloth` package on a given configuration.

    Args:
        config (str): Path to the configuration file
        train_path (str): Path to the training set
        val_path (str): Path to the validation set
        output_dir (str): Path where the model will be saved
        report_to (str): Where to report the results
        seed (int): Random(?) seed
    """

    set_seed(seed)
    logging.set_verbosity_info()

    #
    # Load configuration
    #

    if os.path.exists(config):
        config_path = config
    else:
        import recipes
        recipes_path = os.path.join(recipes.__path__[0])
        config_path = recipes_path + '/configs/' + config + '.yaml'

    # Read config
    with open(config_path, "r") as r:
        config = yaml.safe_load(r)

    # Get paths to train and validation sets
    if train_path is None:
        train_path = config['train_path']
    if val_path is None:
        val_path = config['val_path']

    # Path where model will be saved
    if output_dir is None:
        output_dir = config['output_dir']

    # Get settings of trainer object
    trainer_config = config.get("trainer", {"gradient_checkpointing": False})

    # Get settings of Peft/LoRA adapter
    lora_config = config.get("lora", None)

    # Class to work with Tokenizer
    model_class = "transformers.AutoModelForCausalLM"
    if "class" in config["model"]:
        model_class = config["model"]["class"]

    # Class to work with Tokenizer
    tokenizer_class = "transformers.AutoTokenizer"
    if "class" in config["tokenizer"]:
        tokenizer_class = config["tokenizer"]["class"]

    # Class to work with Tokenizer
    trainer_class = "CustomTrainer"
    if "class" in config["trainer"]:
        trainer_class = config["trainer"]["class"]
        del config["trainer"]["class"]

    # Read repo_id of tokenizer or path to configs on disk#
    tokenizer_name = None
    if "name" in config["tokenizer"]:
        tokenizer_name = config["tokenizer"]["name"]
    elif "name" in config["model"]:
        tokenizer_name = config["model"]["name"]

    # Settings related to bitsandbytes and useful only with LoRA adapter training
    load_in_4bit = False
    if "load_in_4bit" in config["model"]:
        load_in_4bit = bool(config["model"]["load_in_4bit"])
    load_in_8bit = False
    if "load_in_8bit" in config["model"]:
        load_in_8bit = bool(config["model"]["load_in_8bit"])

    #
    # Trainer arguments preparation
    #

    # Get ddp settings from config if available
    ddp_config = config.get("ddp", {})

    # Merge trainer_config and ddp_config
    training_args_dict = trainer_config.copy()
    training_args_dict.update(ddp_config)

    # Fixing "evaL_loss" issue
    training_args_dict.update({"label_names": ["labels"]})

    # If reporting to W&B is enabled
    if report_to == "wandb":
        os.environ["WANDB_MODE"] = "online"
        wandb.init(project="impruver", name=str(config_path), config=config)

    # Prepare trainer configuration
    training_args = UnslothTrainingArguments(
        output_dir=output_dir,
        report_to=report_to,
        run_name=str(config_path),
        **training_args_dict
    )

    #
    # Tokenizer preparation
    #

    # Init tokenizer object
    tokenizer_obj = dynamic_import(tokenizer_class)
    tokenizer = tokenizer_obj.from_pretrained(tokenizer_name)

    # Add special tokens to tokenizer
    if 'special_tokens' in config['tokenizer']:
        special_tokens = {}
        if 'pad_token_id' in config['tokenizer']['special_tokens']:
            special_tokens['pad_token'] = config['tokenizer']['special_tokens']['pad_token']
        if 'eos_token_id' in config['tokenizer']['special_tokens']:
            special_tokens['eos_token'] = config['tokenizer']['special_tokens']['eos_token']
        if 'bos_token_id' in config['tokenizer']['special_tokens']:
            special_tokens['bos_token'] = config['tokenizer']['special_tokens']['bos_token']
        if 'mask_token_id' in config['tokenizer']['special_tokens']:
            special_tokens['mask_token'] = config['tokenizer']['special_tokens']['mask_token']
        if 'unk_token_id' in config['tokenizer']['special_tokens']:
            special_tokens['unk_token'] = config['tokenizer']['special_tokens']['unk_token']
        tokenizer.add_special_tokens(special_tokens)

    # Check if `chat_template` attribute exists in tokenizer
    if not hasattr(tokenizer, 'chat_template') or tokenizer.chat_template is None:
        # If it doesn't exist, use default
        tokenizer.chat_template = DEFAULT_CHAT_TEMPLATE

    # Use provided chat_template if set in config
    if 'chat_template' in config['tokenizer']:
        tokenizer.chat_template = config['tokenizer']['chat_template']

    # Save tokenizer object with all configs to an output folder
    tokenizer.save_pretrained(output_dir)

    #
    # Dataset
    #

    # Read train and evaluation datasets form JSONL
    train_dataset = read_jsonl(train_path)
    val_dataset = read_jsonl(val_path)

    # Randomize order of items in train dataset
    random.shuffle(train_dataset)

    # Init data collator for adding pad tokens to `input_ids` and `labels` lists
    data_collator = DataCollatorForTokenClassification(tokenizer, pad_to_multiple_of=8)

    #
    # Model preparation
    #

    # Data Type is bfloat16 by default
    dtype = torch.bfloat16
    if "dtype" in config["model"]:
        dtype = get_dtype(config["model"]["dtype"])

    # Quantization related settings, can be used only in combination with LoRA adapter
    quantization_config = None
    if load_in_4bit:
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_compute_dtype=dtype,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=True
        )
    if load_in_8bit:
        quantization_config = BitsAndBytesConfig(
            load_in_8bit=True,
            llm_int8_threshold=6.0,
        )
    # If ddp enabled and bnb enabled too then use PartialState
    device_map = None
    if load_in_4bit or load_in_8bit:
        from accelerate import PartialState
        device_map = {"": PartialState().process_index}

    # Attention implementation
    attn_implementation = None
    if "attn_implementation" in config["model"]:
        attn_implementation = config["model"]["attn_implementation"]

    # Init model object
    model_obj = dynamic_import(model_class)

    # If model name is set then pre-train
    if 'name' in config["model"]:
        model = model_obj.from_pretrained(
            config["model"]["name"],
            quantization_config=quantization_config,
            # For DDP without bnb should be None, with bnb need to use PartialState index
            device_map=device_map if ddp_config else "auto",
            torch_dtype=dtype,
            attn_implementation=attn_implementation
        )
    else:
        # Init from scratch
        model_config_class = config["model"]['config_class']
        model_config_class_obj = dynamic_import(model_config_class)
        model_config = model_config_class_obj(**config["model"]["config"])
        model = model_obj._from_config(model_config)

    # If we need to train a LoRA adapter
    if lora_config:
        lora_config = LoraConfig(**lora_config)
        model = get_peft_model(model, lora_config)
        model.print_trainable_parameters()

        # First we need to check if we have a default generation config in model
        if hasattr(model, "generation_config") and model.generation_config is not None:
            generation_config = model.generation_config.to_diff_dict()
        else:
            generation_config = {"max_new_tokens": 200, "repetition_penalty": 1.2, "do_sample": True}
        # Next let's save a default generation config
        generation_config_path = os.path.join(output_dir, "generation_config.json")
        with open(generation_config_path, "w", encoding="utf-8") as f:
            json.dump(generation_config, f, ensure_ascii=False, default=lambda o: o.__dict__, indent=2)

    #
    # Trainer
    #

    # Load trainer class object
    trainer_obj = CustomTrainer
    if trainer_class != "CustomTrainer":
        trainer_obj = dynamic_import(trainer_class)

    # Init trainer object and pass all important parameters to it
    trainer = trainer_obj(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=val_dataset,
        data_collator=data_collator,
    )

    #
    # Training loop
    #

    trainer.train()

    #
    # Saving results
    #

    model.save_pretrained(output_dir, safe_serialization=False)


if __name__ == "__main__":
    fire.Fire(finetune)



from typing import Dict, Optional

import torch

PRECISION_STR_TO_DTYPE: Dict[str, torch.dtype] = {
    "fp16": torch.float16,
    "bf16": torch.bfloat16,
    "fp32": torch.float32,
    "fp64": torch.float64,
}


def verify_bf16_support() -> bool:
    return (
            torch.cuda.is_available()
            and torch.cuda.is_bf16_supported()
            and torch.distributed.is_nccl_available()
            and torch.cuda.nccl.version() >= (2, 10)
    )


def get_dtype(dtype: Optional[str] = None, device: Optional[torch.device] = None) -> torch.dtype:
    # None defaults to float32
    if dtype is None:
        return torch.float32

    # Convert to torch.dtype
    torch_dtype = PRECISION_STR_TO_DTYPE.get(dtype, dtype)

    # dtype must be one of the supported precisions
    if torch_dtype not in PRECISION_STR_TO_DTYPE.values():
        raise ValueError(
            f"Dtype {torch_dtype} must be one of {', '.join(list(PRECISION_STR_TO_DTYPE.keys()))} for finetuning."
        )

    if torch_dtype == torch.bfloat16 and device != torch.device("cpu") and not verify_bf16_support():
        raise RuntimeError(
            "bf16 precision was requested but not available on this hardware. Please use fp32 precision instead."
        )

    return torch_dtype



def dialog_to_messages(sample: dict) -> list | None:
    """
    Converts a dialog structure into a list of messages compatible with the OpenAI format.

    Example:
        {"messages": {"roles": ["user", "assistant"], "content": ["user text", "assistant text"]}}

    Args:
        sample (dict): a dictionary containing the dialog data

    Returns:
        list: a list of messages compatible with the OpenAI format
        [{"role": "user", "content": "user text"}, {"role": "assistant", "content": "assistant text"}]
    """
    messages = []
    roles = sample['messages']['role']
    contents = sample['messages']['content']
    if (len(roles) != len(contents)) or (len(roles) < 2):
        return None
    for role, content in zip(roles, contents):
        messages.append({
            "role": 'user' if role == 'user' else 'assistant',
            "content": content
        })
    return messages



import torch
from torch.utils.data import Dataset
from typing import Callable, Dict, Any, Optional, List
from datasets import load_dataset

class RawDataset(Dataset):
    def __init__(
        self,
        *,
        tokenizer,
        source: str,
        max_tokens_count: Optional[int] = None,
        convert_function: Optional[Callable[[Dict[str, Any]], str]] = None,
        strategy_function: Optional[Callable[[str, Any], str]] = None,
        **load_dataset_kwargs: Any,
    ):
        self.tokenizer = tokenizer
        self.source = source
        self.convert_function = convert_function
        self.strategy_function = strategy_function
        self.data = load_dataset(self.source, **load_dataset_kwargs)
        self.max_tokens_count = max_tokens_count or tokenizer.model_max_length

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        sample = self.data[index]
        text = self.convert_function(sample) if self.convert_function else sample['text']

        # Apply the strategy function if provided
        if self.strategy_function is not None:
            text = self.strategy_function(
                text=text,
                tokenizer=self.tokenizer,
                max_tokens_count=self.max_tokens_count,
            )

        # Tokenize the text
        tokenized = self.tokenizer(
            text,
            max_length=self.max_tokens_count,
            truncation=True,
            padding='max_length',
            return_tensors='pt',
        )

        input_ids = tokenized['input_ids'].squeeze(0)
        attention_mask = tokenized['attention_mask'].squeeze(0)

        return {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
        }



def instruction_to_messages(sample: dict, skip_labels: list = ["bad_output"], mapping: dict = None) -> list | None:
    """
    Convert a dataset sample to a list of messages with optional field mapping.

    Args:
        sample (dict): Dataset sample with keys like {"instruction": "text", "input": "text", "output": "text"}.
        skip_labels (list, optional): List of labels to skip, defaults to ["bad_output"].
        mapping (dict, optional): Mapping for field names, e.g., {"instruction": "question", "output": "answer"}.

    Returns:
        list | None: List of messages compatible with OpenAI format, or None if sample is skipped.
    """
    # Apply mapping or use defaults
    instruction_key = mapping.get("instruction", "instruction") if mapping else "instruction"
    input_key = mapping.get("input", "input") if mapping else "input"
    output_key = mapping.get("output", "output") if mapping else "output"
    system_key = mapping.get("system", None) if mapping else None

    if "label" in sample and sample.get(input_key) in skip_labels:
        return None

    instruction = sample.get(instruction_key, "")
    if input_key in sample and sample[input_key]:
        instruction += "\n" + sample[input_key]

    messages = [
        {"role": "user", "content": instruction},
        {"role": "assistant", "content": sample.get(output_key, "")}
    ]

    if system_key and system_key in sample:
        messages.insert(0, {"role": "system", "content": sample.get(system_key, "")})

    return messages



from dataclasses import dataclass
from typing import Literal, Dict

Role = Literal[
    "system",
    "user",
    "think",
    "functioncall",
    "functionresponse",
    "assistant"
]


@dataclass
class Message:
    """
    Represents a siungle message in a conversation.
    A message can be a text message, a function call, or a function response.
    """

    role: Role
    content: str

    @classmethod
    def from_dict(cls, d: dict)-> "Message":
        """
        Construct a Message object from a dictionary.

        Args:
            d (dict): A dictionary containing the role and content of the message.

        Returns:
            Message: A Message object.
        """
        return cls(role=d["role"], content=d["content"])

    def to_dict(self) -> Dict[str, str]:
        """
        Convert the Message object to a dictionary.

        Returns:
            dict: A dictionary containing the role and content of the message.
        """
        return {"role": self.role, "content": self.content}



import logging
from transformers import logging as t_logging


def get_logger(level=t_logging.log_levels['info']) -> logging.Logger:
    t_logging.set_verbosity(level)
    logger = t_logging.get_logger(__file__)
    logger.setLevel(level)
    return logger



import argparse

from impruver._cli.download import Download
from impruver._cli.copy import Copy
from impruver._cli.list import List
from impruver._cli.run import Run


class Parser:
    """Holds all information related to running the CLI"""

    def __init__(self):
        # Initialize the top-level parser
        self._parser = argparse.ArgumentParser(
            prog="impruver",
            description="Welcome to the Impruver!",
            add_help=True,
        )

        # Default command is to print help
        self._parser.set_defaults(func=lambda args: self._parser.print_help())

        # Add subcommands
        subparsers = self._parser.add_subparsers(title="subcommands")
        Download.create(subparsers)
        List.create(subparsers)
        Copy.create(subparsers)
        Run.create(subparsers)

    def parse_args(self) -> argparse.Namespace:
        """Parse CLI arguments"""
        return self._parser.parse_args()

    def run(self, args: argparse.Namespace) -> None:
        """Execute CLI"""
        args.func(args)


def main():
    parser = Parser()
    args = parser.parse_args()
    parser.run(args)


if __name__ == "__main__":
    main()



import argparse
import shutil
import textwrap
from pathlib import Path

import impruver
from impruver._cli.subcommand import Subcommand
from impruver._recipe_registry import get_all_recipes

ROOT = Path(impruver.__file__).parent.parent


class Copy(Subcommand):
    """Содержит всю логику для подкоманды `tune cp`."""

    def __init__(self, subparsers):
        super().__init__()
        self._parser = subparsers.add_parser(
            "cp",
            prog="impruver cp",
            usage="impruver cp <recipe|config> <destination> [OPTIONS]",
            help="Copy a built-in recipe or config to a local path.",
            description="Copy a built-in recipe or config to a local path.",
            epilog=textwrap.dedent(
                """\
            examples:
                $ impruver cp finetune .
                Copied file to ./finetune_transformers.py

                $ impruver cp ruGPT-3.5/13B_lora_saiga2 ./new_dir/my_custom_lora.yaml --make-parents
                Copied file to ./new_dir/my_custom_lora.yaml

            Need to see all possible recipes/configs to copy? Try running `impruver ls`.
            """
            ),
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self._add_arguments()
        self._parser.set_defaults(func=self._cp_cmd)

    def _add_arguments(self) -> None:
        self._parser.add_argument(
            "file", type=str,
            help="Recipe/config to copy. For a list of all possible options, run `impruver ls`",
        )
        self._parser.add_argument(
            "destination", type=Path,
            help="Location to copy the file to",
        )
        self._parser.add_argument(
            "-n", "--no-clobber", action="store_true", default=False,
            help="Do not overwrite destination if it already exists",

        )
        self._parser.add_argument(
            "--make-parents", action="store_true", default=False,
            help="Create parent directories for destination if they do not exist. "
                 "If not set to True, will error if parent directories do not exist",
        )

    def _cp_cmd(self, args: argparse.Namespace):
        destination: Path = args.destination
        src = None

        # Iterate through all recipes and configs
        for recipe in get_all_recipes():
            if recipe.name == args.file:
                src = ROOT / "recipes" / recipe.file_path
                proper_suffix = ".py"
                break
            for config in recipe.configs:
                if config.name == args.file:
                    src = ROOT / "recipes" / "configs" / config.file_path
                    proper_suffix = ".yaml"
                    break

        # Fail if no file exists
        if src is None:
            self._parser.error(
                f"Invalid file name: {args.file}. Try `impruver ls` to see all available files to copy."
            )

        # Attach proper suffix if needed
        if destination.name != "" and destination.suffix != proper_suffix:
            destination = destination.with_suffix(proper_suffix)

        # Copy file
        try:
            if args.no_clobber and destination.exists():
                print(
                    f"File already exists at {destination.absolute()}, not overwriting."
                )
            else:
                if args.make_parents:
                    destination.parent.mkdir(parents=True, exist_ok=True)
                output = shutil.copy(src, destination)
                print(f"Copied file to {output}")
        except FileNotFoundError:
            self._parser.error(
                f"Cannot create regular file: '{destination}'. No such file or directory. "
                "If the specified destination's parent directory does not exist and you would "
                "like to create it on-the-fly, use the --make-parents flag."
            )



def generate(model, tokenizer, prompt, generation_config):
    data = tokenizer(prompt, return_tensors="pt")
    data = {k: v.to(model.device) for k, v in data.items()}
    output_ids = model.generate(**data, generation_config=generation_config)[0]
    output_ids = output_ids[len(data["input_ids"][0]):]
    output = tokenizer.decode(output_ids, skip_special_tokens=True)
    return output.strip()



import argparse
import textwrap

from impruver._cli.subcommand import Subcommand
from impruver._recipe_registry import get_all_recipes


class List(Subcommand):
    """Holds all the logic for the `tune ls` subcommand."""

    NULL_VALUE = "<любой>"

    def __init__(self, subparsers: argparse._SubParsersAction):
        super().__init__()
        self._parser = subparsers.add_parser(
            "ls",
            prog="impruver ls",
            help="List all built-in recipes and configs",
            description="List all built-in recipes and configs",
            epilog=textwrap.dedent(
                """\
            examples:
                $ impruver ls
                RECIPE               CONFIG
                finetune             ruGPT-3.5/13B_lora_saiga2
                                     rugpt3large/760M_full_fc
                                     rugpt3large/760M_full_saiga2
                                     zephyr/7B_lora_python-ru
                compose_dataset      <любой>
                chat                 <любой>
                ...

            To run one of these recipes:
                $ impruver run finetune --config ruGPT-3.5/13B_lora_saiga2
            """
            ),
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self._parser.set_defaults(func=self._ls_cmd)

    def _ls_cmd(self, args: argparse.Namespace) -> None:
        """List all available recipes and configs."""
        # Print table header
        header = f"{'RECIPE':<20} {'CONFIG':<40}"
        print(header)

        # Print recipe/config pairs
        for recipe in get_all_recipes():
            # If there are no configs for a recipe, print a blank config
            recipe_str = recipe.name
            if len(recipe.configs) == 0:
                row = f"{recipe_str:<20} {self.NULL_VALUE:<40}"
                print(row)
            for i, config in enumerate(recipe.configs):
                # If there are multiple configs for a single recipe, omit the recipe name
                # on latter configs
                if i > 0:
                    recipe_str = ""
                row = f"{recipe_str:<20} {config.name:<40}"
                print(row)



from .message import Message
from .tokenizer import Tokenizer
from .strategies import last_message_by_assistant
from .validate_messages import validate_messages
from .apply_chat_template import apply_chat_template, DEFAULT_CHAT_TEMPLATE
